import os
import logging
import shutil
from tempfile import mkstemp

from larigira.fsutils import shortname

log = logging.getLogger(__name__)


def generate(spec):
    """
    resolves audiospec-static

    Recognized argument is  "paths" (list of static paths)
    """
    if "paths" not in spec:
        raise ValueError("Malformed audiospec: missing 'paths'")

    for path in spec["paths"]:
        if not os.path.exists(path):
            log.warning("Can't find requested path: %s", path)
            continue
        tmp = mkstemp(
            suffix=os.path.splitext(path)[-1], prefix="static-%s-" % shortname(path)
        )
        os.close(tmp[0])
        log.info("copying %s -> %s", path, os.path.basename(tmp[1]))
        shutil.copy(path, tmp[1])
        yield "file://{}".format(tmp[1])


generate.description = "Picks always the same specified file"



from __future__ import print_function
import sys
import argparse
from .entrypoints_utils import get_one_entrypoint
import json
from logging import getLogger

log = getLogger("audiogen")


def get_audiogenerator(kind):
    """Messes with entrypoints to return an audiogenerator function"""
    return get_one_entrypoint("larigira.audiogenerators", kind)


def get_parser():
    parser = argparse.ArgumentParser(description="Generate audio and output paths")
    parser.add_argument(
        "audiospec",
        metavar="AUDIOSPEC",
        type=str,
        nargs=1,
        help="filename for audiospec, formatted in json",
    )
    return parser


def read_spec(fname):
    try:
        if fname == "-":
            return json.load(sys.stdin)
        with open(fname) as buf:
            return json.load(buf)
    except ValueError:
        sys.stderr.write("Error: invalid JSON\n")
        sys.exit(1)


def check_spec(spec):
    if "kind" not in spec:
        yield "Missing field 'kind'"


def audiogenerate(spec):
    gen = get_audiogenerator(spec["kind"])
    return tuple(gen(spec))


def main():
    """Main function for the "larigira-audiogen" executable"""
    args = get_parser().parse_args()
    spec = read_spec(args.audiospec[0])
    errors = tuple(check_spec(spec))
    if errors:
        log.error("Errors in audiospec")
        for err in errors:
            sys.stderr.write("Error: {}\n".format(err))
        sys.exit(1)
    for path in audiogenerate(spec):
        print(path)


if __name__ == "__main__":
    main()



import logging

from flask_wtf import Form
from wtforms import StringField, TextAreaField, validators, SubmitField, ValidationError
from croniter import croniter

log = logging.getLogger(__name__)


class CronAlarmForm(Form):
    nick = StringField(
        "Alarm nick",
        validators=[validators.required()],
        description="A simple name to recognize this alarm",
    )
    cron_format = StringField(
        "cron-like format",
        validators=[validators.required()],
        description="the frequency specification, as in the <tt>cron</tt> command; "
        'see <a href="https://crontab.guru/">crontab.guru</a> for a hepl with cron format',
    )
    exclude = TextAreaField(
        "cron-like format; any matching time will be excluded",
        description="Another cron-like thing to _exclude_ events",
    )
    submit = SubmitField("Submit")

    def populate_from_timespec(self, timespec):
        if "nick" in timespec:
            self.nick.data = timespec["nick"]
        if "cron_format" in timespec:
            self.cron_format.data = timespec["cron_format"]
        if "exclude" in timespec:
            if type(timespec["exclude"]) is str:
                self.exclude.data = timespec["exclude"]
            else:
                self.exclude.data = "\n".join(timespec["exclude"])

    def validate_cron_format(self, field):
        if not croniter.is_valid(field.data):
            raise ValidationError("formato di cron non valido")

    def validate_exclude(self, field):
        for line in field.data.split("\n"):
            if line.strip() and not croniter.is_valid(line):
                raise ValidationError("formato di exclude non valido: %s" % line)


def cronalarm_receive(form):
    return {
        "kind": "cron",
        "nick": form.nick.data,
        "cron_format": form.cron_format.data,
        "exclude": [
            line.strip() for line in form.exclude.data.split("\n") if line.strip()
        ],
    }



import wave


def maxwait(songs, context, conf):
    wait = int(conf.get("EF_MAXWAIT_SEC", 0))
    if wait == 0:
        return True
    if "time" not in context["status"]:
        return True, "no song playing?"
    curpos, duration = map(int, context["status"]["time"].split(":"))
    remaining = duration - curpos
    if remaining > wait:
        return False, "remaining %d max allowed %d" % (remaining, wait)
    return True


def get_duration(path):
    """get track duration in seconds"""
    if path.lower().endswith(".wav"):
        with wave.open(path, "r") as f:
            frames = f.getnframes()
            rate = f.getframerate()
            duration = frames / rate
            return int(duration)
    try:
        import mutagen
    except ModuleNotFoundError:
        raise ImportError("mutagen not installed")

    audio = mutagen.File(path)
    if audio is None:
        return None
    return int(audio.info.length)


def percentwait(songs, context, conf, getdur=get_duration):
    """
    Similar to maxwait, but the maximum waiting time is proportional to the
    duration of the audio we're going to add.

    This filter observe the EF_MAXWAIT_PERC variable.
    The variable must be an integer representing the percentual.
    If the variable is 0 or unset, the filter will not run.

    For example, if the currently playing track still has 1 minute to go and we
    are adding a jingle of 40seconds, then if EF_MAXWAIT_PERC==200 the audio
    will be added (40s*200% = 1m20s) while if EF_MAXWAIT_PERC==100 it will be
    filtered out.
    """
    percentwait = int(conf.get("EF_MAXWAIT_PERC", 0))
    if percentwait == 0:
        return True
    if "time" not in context["status"]:
        return True, "no song playing?"
    curpos, duration = map(int, context["status"]["time"].split(":"))
    remaining = duration - curpos
    eventduration = 0
    for uri in songs["uris"]:
        if not uri.startswith("file://"):
            return True, "%s is not a file" % uri
        path = uri[len("file://") :]  # strips file://
        songduration = getdur(path)
        if songduration is None:
            continue
        eventduration += songduration

    if eventduration == 0:
        # must be an error! mutagen support is not always perfect
        return True, 'mutagen could not calculate length of %s' % uri
    wait = eventduration * (percentwait / 100.0)
    if remaining > wait:
        return False, "remaining %d max allowed %d" % (remaining, wait)
    return True



import os
import logging
import shutil
import time
from tempfile import mkstemp

from pytimeparse.timeparse import timeparse

from larigira.fsutils import scan_dir, shortname

log = logging.getLogger(__name__)


def get_mtime(fname):
    return int(os.path.getmtime(fname))


def recent_choose(paths, howmany, minepoch):
    found_files = {}
    for path in paths:
        if not os.path.exists(path):
            log.warning("Can't find requested path: %s", path)
            continue
        if os.path.isfile(path):
            found_files[path] = get_mtime(path)
        elif os.path.isdir(path):
            found_files.update({fname: get_mtime(fname) for fname in scan_dir(path)})
    found_files = [
        (fname, mtime) for (fname, mtime) in found_files.items() if mtime >= minepoch
    ]

    return [
        fname
        for fname, mtime in sorted(found_files, key=lambda x: x[1], reverse=True)[
            :howmany
        ]
    ]


def generate(spec):
    """
    resolves audiospec-randomdir

    Recognized arguments:
        - path [mandatory]    source dir
        - maxage [default=ignored]  max age of audio files to pick
    """
    for attr in ("path", "maxage"):
        if attr not in spec:
            raise ValueError("Malformed audiospec: missing '%s'" % attr)

    if spec["maxage"].strip():
        try:
            maxage = int(spec["maxage"])
        except ValueError:
            maxage = timeparse(spec["maxage"])
            if maxage is None:
                raise ValueError(
                    "Unknown format for maxage: '{}'".format(spec["maxage"])
                )
            assert type(maxage) is int
    else:
        maxage = None

    now = int(time.time())
    minepoch = 0 if maxage is None else now - maxage

    picked = recent_choose([spec["path"]], 1, minepoch)

    for path in picked:
        tmp = mkstemp(
            suffix=os.path.splitext(path)[-1], prefix="randomdir-%s-" % shortname(path)
        )
        os.close(tmp[0])
        shutil.copy(path, tmp[1])
        log.info("copying %s -> %s", path, os.path.basename(tmp[1]))
        yield "file://{}".format(tmp[1])


generate.description = "Select most recent file inside a directory"



import flask_wtf
from wtforms import StringField, validators, SubmitField, IntegerField

from larigira.formutils import AutocompleteStringField


class Form(flask_wtf.Form):
    nick = StringField(
        "Audio nick",
        validators=[validators.required()],
        description="A simple name to recognize this audio",
    )
    path = AutocompleteStringField(
        "dl-suggested-dirs",
        "Path",
        validators=[validators.required()],
        description="Full path to source directory",
    )
    howmany = IntegerField(
        "Number",
        validators=[validators.optional()],
        default=1,
        description="How many songs to be picked" "from this dir; defaults to 1",
    )
    submit = SubmitField("Submit")

    def populate_from_audiospec(self, audiospec):
        if "nick" in audiospec:
            self.nick.data = audiospec["nick"]
        if "paths" in audiospec:
            self.path.data = audiospec["paths"][0]
        if "howmany" in audiospec:
            self.howmany.data = audiospec["howmany"]
        else:
            self.howmany.data = 1


def receive(form):
    return {
        "kind": "randomdir",
        "nick": form.nick.data,
        "paths": [form.path.data],
        "howmany": form.howmany.data or 1,
    }



from pytimeparse.timeparse import timeparse
from flask_wtf import Form
from wtforms import StringField, validators, SubmitField, ValidationError

from larigira.formutils import AutocompleteStringField


class AudioForm(Form):
    nick = StringField(
        "Audio nick",
        validators=[validators.required()],
        description="A simple name to recognize this audio",
    )
    path = AutocompleteStringField(
        "dl-suggested-dirs",
        "Path",
        validators=[validators.required()],
        description="Directory to pick file from",
    )
    maxage = StringField(
        "Max age",
        validators=[validators.required()],
        description="in seconds, or human-readable " "(like 9w3d12h)",
    )
    submit = SubmitField("Submit")

    def validate_maxage(self, field):
        try:
            int(field.data)
        except ValueError:
            if timeparse(field.data) is None:
                raise ValidationError(
                    "maxage must either be a number "
                    "(in seconds) or a human-readable "
                    "string like '1h2m'  or '1d12h'"
                )

    def populate_from_audiospec(self, audiospec):
        if "nick" in audiospec:
            self.nick.data = audiospec["nick"]
        if "path" in audiospec:
            self.path.data = audiospec["path"]
        if "maxage" in audiospec:
            self.maxage.data = audiospec["maxage"]


def audio_receive(form):
    return {
        "kind": "mostrecent",
        "nick": form.nick.data,
        "path": form.path.data,
        "maxage": form.maxage.data,
        "howmany": 1,
    }



from flask_wtf import Form
from wtforms import StringField, validators, SubmitField, ValidationError

from larigira.formutils import AutocompleteStringField


class ScriptAudioForm(Form):
    nick = StringField(
        "Audio nick",
        validators=[validators.required()],
        description="A simple name to recognize this audio",
    )
    name = AutocompleteStringField(
        "dl-suggested-scripts",
        "Name",
        validators=[validators.required()],
        description="filename (NOT path) of the script",
    )
    args = StringField("Arguments", description='arguments, separated by ";"')
    submit = SubmitField("Submit")

    def populate_from_audiospec(self, audiospec):
        if "nick" in audiospec:
            self.nick.data = audiospec["nick"]
        if "name" in audiospec:
            self.name.data = audiospec["name"]
        if "args" in audiospec:
            if type(audiospec["args"]) is str:  # legacy compatibility
                self.args.data = audiospec["args"].replace(" ", ";")
            else:
                self.args.data = ";".join(audiospec["args"])

    def validate_name(self, field):
        if "/" in field.data:
            raise ValidationError(
                "Name cannot have slashes: " "it's a name, not a path"
            )


def scriptaudio_receive(form):
    return {
        "kind": "script",
        "nick": form.nick.data,
        "name": form.name.data,
        "args": form.args.data.split(";"),
    }



import os
import fnmatch
import mimetypes


def scan_dir(dirname, extension=None):
    if extension is None:
        extension = "*"
    for root, dirnames, filenames in os.walk(dirname):
        for fname in fnmatch.filter(filenames, extension):
            yield os.path.join(root, fname)


def multi_fnmatch(fname, extensions):
    for ext in extensions:
        if fnmatch.fnmatch(fname, "*." + ext):
            return True
    return False


def is_audio(fname):
    mimetype = mimetypes.guess_type(fname)[0]
    if mimetype is None:
        return False
    return mimetype.split("/")[0] == "audio"


def scan_dir_audio(dirname, extensions=("mp3", "oga", "wav", "ogg")):
    for root, dirnames, filenames in os.walk(dirname):
        for fname in filenames:
            if is_audio(fname):
                yield os.path.join(root, fname)


def shortname(path):
    name = os.path.basename(path)  # filename
    name = name.rsplit(".", 1)[0]  # no extension
    name = "".join(c for c in name if c.isalnum())  # no strange chars
    return name



import logging

from wtforms.fields import StringField
from wtforms import Field
import wtforms.widgets

from datetime import datetime

log = logging.getLogger(__name__)


class AutocompleteTextInput(wtforms.widgets.Input):
    def __init__(self, datalist=None):
        super().__init__("text")
        self.datalist = datalist

    def __call__(self, field, **kwargs):
        # every second can be specified
        if self.datalist is not None:
            return super(AutocompleteTextInput, self).__call__(
                field, list=self.datalist, autocomplete="autocomplete", **kwargs
            )
        return super(AutocompleteTextInput, self).__call__(field, **kwargs)


class AutocompleteStringField(StringField):
    def __init__(self, datalist, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget = AutocompleteTextInput(datalist)


class DateTimeInput(wtforms.widgets.Input):
    input_type = "datetime-local"

    def __call__(self, field, **kwargs):
        # every second can be specified
        return super(DateTimeInput, self).__call__(field, step="1", **kwargs)


class EasyDateTimeField(Field):
    """
    a "fork" of DateTimeField which uses HTML5 datetime-local

    The format is not customizable, because it is imposed by the HTML5
    specification.

    This field does not ensure that browser actually supports datetime-local
    input type, nor does it provide polyfills.
    """

    widget = DateTimeInput()
    formats = ("%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M")

    def __init__(self, label=None, validators=None, **kwargs):
        super(EasyDateTimeField, self).__init__(label, validators, **kwargs)

    def _value(self):
        if self.raw_data:
            return " ".join(self.raw_data)
        return self.data and self.data.strftime(self.formats[0]) or ""

    def process_formdata(self, valuelist):
        if valuelist:
            date_str = " ".join(valuelist)
            for fmt in self.formats:
                try:
                    self.data = datetime.strptime(date_str, fmt)
                    return
                except ValueError:
                    log.debug("Format `%s` not valid for `%s`", fmt, date_str)
            raise ValueError(
                self.gettext("Not a valid datetime value <tt>{}</tt>").format(date_str)
            )



"""
This component will look for files to be removed. There are some assumptions:
    * Only files in $TMPDIR are removed. Please remember that larigira has its
      own specific TMPDIR
    * MPD URIs are parsed, and only file:/// is supported
"""
import os
from os.path import normpath
import logging
import mpd


def old_commonpath(directories):
    if any(p for p in directories if p.startswith("/")) and any(
        p for p in directories if not p.startswith("/")
    ):
        raise ValueError("Can't mix absolute and relative paths")
    norm_paths = [normpath(p) + os.path.sep for p in directories]
    ret = os.path.dirname(os.path.commonprefix(norm_paths))
    if len(ret) > 0 and ret == "/" * len(ret):
        return "/"
    return ret


try:
    from os.path import commonpath
except ImportError:
    commonpath = old_commonpath


class UnusedCleaner:
    def __init__(self, conf):
        self.conf = conf
        self.waiting_removal_files = set()
        self.log = logging.getLogger(self.__class__.__name__)

    def _get_mpd(self):
        mpd_client = mpd.MPDClient(use_unicode=True)
        mpd_client.connect(self.conf["MPD_HOST"], self.conf["MPD_PORT"])
        return mpd_client

    def watch(self, uri):
        """
        adds fpath to the list of "watched" file

        as soon as it leaves the mpc playlist, it is removed
        """
        if not uri.startswith("file:///"):
            return  # not a file URI
        fpath = uri[len("file://") :]
        if (
            "TMPDIR" in self.conf
            and self.conf["TMPDIR"]
            and commonpath([self.conf["TMPDIR"], fpath])
            != normpath(self.conf["TMPDIR"])
        ):
            self.log.info("Not watching file %s: not in TMPDIR", fpath)
            return
        if not os.path.exists(fpath):
            self.log.warning("a path that does not exist is being monitored")
        self.waiting_removal_files.add(fpath)

    def check_playlist(self):
        """check playlist + internal watchlist to see what can be removed"""
        mpdc = self._get_mpd()
        files_in_playlist = {
            song["file"] for song in mpdc.playlistid() if song["file"].startswith("/")
        }
        for fpath in self.waiting_removal_files - files_in_playlist:
            # we can remove it!
            self.log.debug("removing unused: %s", fpath)
            self.waiting_removal_files.remove(fpath)
            if os.path.exists(fpath):
                os.unlink(fpath)



import gevent


class ParentedLet(gevent.Greenlet):
    """
    ParentedLet is just a helper subclass that will help you when your
    greenlet main duty is to "signal" things to a parent_queue.

    It won't save you much code, but "standardize" messages and make explicit
    the role of that greenlet
    """

    def __init__(self, queue):
        gevent.Greenlet.__init__(self)
        self.parent_queue = queue
        self.parent_greenlet = None
        self.tracker = None  # set this to recognize easily

    def parent_msg(self, kind, *args):
        return {
            "emitter": self,
            "class": self.__class__.__name__,
            "tracker": self.tracker,
            "kind": kind,
            "args": args,
        }

    def send_to_parent(self, kind, *args):
        self.parent_queue.put(self.parent_msg(kind, *args))

    def _run(self):
        if not hasattr(self, "do_business"):
            raise Exception(
                "do_business method not implemented by %s" % self.__class__.__name__
            )
        for msg in self.do_business():
            self.send_to_parent(*msg)


class Timer(ParentedLet):
    """continously sleeps some time, then send a "timer" message to parent"""

    def __init__(self, milliseconds, queue):
        ParentedLet.__init__(self, queue)
        self.ms = milliseconds

    def parent_msg(self, kind, *args):
        msg = ParentedLet.parent_msg(self, kind, *args)
        msg["period"] = self.ms
        return msg

    def do_business(self):
        while True:
            gevent.sleep(self.ms / 1000.0)
            yield ("timer",)



from __future__ import print_function
import sys
import argparse
import json

from .db import EventModel
from .config import get_conf


def main_list(args):
    m = EventModel(args.file)
    for alarm, action in m.get_all_alarms_expanded():
        json.dump(dict(alarm=alarm, action=action), sys.stdout, indent=4)
        sys.stdout.write('\n')


def main_getaction(args):
    m = EventModel(args.file)
    json.dump(m.get_action_by_id(int(args.actionid)), sys.stdout, indent=4)


def main_add(args):
    m = EventModel(args.file)
    m.add_event(
        dict(kind="frequency", interval=args.interval, start=1),
        [dict(kind="mpd", howmany=1)],
    )


def main():
    conf = get_conf()
    p = argparse.ArgumentParser()
    p.set_defaults(func=None)
    p.add_argument(
        "-f", "--file", help="Filepath for DB", required=False, default=conf["DB_URI"]
    )
    sub = p.add_subparsers()
    sub_list = sub.add_parser("list")
    sub_list.set_defaults(func=main_list)

    sub_getaction = sub.add_parser("getaction")
    sub_getaction.set_defaults(func=main_getaction)
    sub_getaction.add_argument('actionid')

    sub_add = sub.add_parser("add")
    sub_add.add_argument("--interval", type=int, default=3600)
    sub_add.set_defaults(func=main_add)

    args = p.parse_args()
    if args.func is None:
        p.print_help()
        sys.exit(2)
    args.func(args)


if __name__ == "__main__":
    main()



from __future__ import print_function
import logging

log = logging.getLogger("time-every")
from datetime import datetime, timedelta

from pytimeparse.timeparse import timeparse


def getdate(val):
    if type(val) is int:
        return datetime.fromtimestamp(val)
    return val


class Alarm(object):
    def __init__(self):
        pass

    def next_ring(self, current_time=None):
        """if current_time is None, it is now()

        returns the next time it will ring; or None if it will not anymore
        """
        raise NotImplementedError()

    def has_ring(self, time=None):
        """returns True IFF the alarm will ring exactly at ``time``"""
        raise NotImplementedError()

    def all_rings(self, current_time=None):
        """
        all future rings
        this, of course, is an iterator (they could be infinite)
        """
        ring = self.next_ring(current_time)
        while ring is not None:
            yield ring
            ring = self.next_ring(ring)


class SingleAlarm(Alarm):
    """
    rings a single time
    """

    description = "Only once, at a specified date and time"

    def __init__(self, obj):
        super().__init__()
        self.dt = getdate(obj["timestamp"])

    def next_ring(self, current_time=None):
        """if current_time is None, it is now()"""
        if current_time is None:
            current_time = datetime.now()
        if current_time >= self.dt:
            return None
        return self.dt

    def has_ring(self, current_time=None):
        if current_time is None:
            current_time = datetime.now()
        return current_time == self.dt


class FrequencyAlarm(Alarm):
    """
    rings on {t | exists a k integer >= 0 s.t. t = start+k*t, start<t<end}
    """

    description = "Events at a specified frequency. Example: every 30minutes"

    def __init__(self, obj):
        self.start = getdate(obj["start"])
        try:
            self.interval = int(obj["interval"])
        except ValueError:
            self.interval = timeparse(obj["interval"])
        assert type(self.interval) is int
        self.end = getdate(obj["end"]) if "end" in obj else None
        self.weekdays = [int(x) for x in obj["weekdays"]] if "weekdays" in obj else None
        if self.weekdays is not None:
            for weekday in self.weekdays:
                if not 1 <= weekday <= 7:
                    raise ValueError("Not a valid weekday: {}".format(weekday))

    def next_ring(self, current_time=None):
        """if current_time is None, it is now()"""
        if current_time is None:
            current_time = datetime.now()
        if self.end is not None and current_time > self.end:
            return None
        if current_time < self.start:
            return self.start
        if self.end is not None:
            assert self.start <= current_time <= self.end
        else:
            assert self.start <= current_time
        # this "infinite" loop is required by the weekday exclusion: in
        # fact, it is necessary to retry until a valid event/weekday is
        # found. a "while True" might have been more elegant (and maybe
        # fast), but this gives a clear upper bound to the cycle.
        for _ in range(max(60 * 60 * 24 * 7 // self.interval, 1)):
            n_interval = (
                (current_time - self.start).total_seconds() // self.interval
            ) + 1
            ring = self.start + timedelta(seconds=self.interval * n_interval)
            if ring == current_time:
                ring += timedelta(seconds=self.interval)
            if self.end is not None and ring > self.end:
                return None
            if self.weekdays is not None and ring.isoweekday() not in self.weekdays:
                current_time = ring
                continue
            return ring
        log.warning(
            "Can't find a valid time for event %s; " "something went wrong", str(self)
        )
        return None

    def has_ring(self, current_time=None):
        if current_time is None:
            current_time = datetime.now()
        if not self.start >= current_time >= self.end:
            return False

        n_interval = (current_time - self.start).total_seconds() // self.interval
        expected_time = self.start + timedelta(seconds=self.interval * n_interval)
        return expected_time == current_time

    def __str__(self):
        return "FrequencyAlarm(every %ds)" % self.interval



#!/usr/bin/env python
"""
Taken from flask-appconfig
"""

import json
import os

from xdg import BaseDirectory


def get_conf(prefix="LARIGIRA_"):
    """This is where everyone should get configuration from"""
    conf_dir = BaseDirectory.save_config_path("larigira")
    conf = {}
    conf["CONTINOUS_AUDIOSPEC"] = dict(kind="mpd", howmany=1)
    conf["MPD_HOST"] = os.getenv("MPD_HOST", "localhost")
    conf["MPD_PORT"] = int(os.getenv("MPD_PORT", "6600"))
    conf["CACHING_TIME"] = 10
    conf["DB_URI"] = os.path.join(conf_dir, "db.json")
    conf["SCRIPTS_PATH"] = os.path.join(conf_dir, "scripts")
    conf["EXTRA_STATIC_PATH"] = os.path.join(conf_dir, "extra")
    conf["EXTRA_MENU_LINKS"] = []
    conf["ROUTE_PREFIX"] = ""
    conf["BOOTSTRAP_SERVE_LOCAL"] = True
    conf["SECRET_KEY"] = "Please replace me!"
    conf["MPD_WAIT_START"] = True
    conf["MPD_WAIT_START_RETRYSECS"] = 5
    conf["CHECK_SECS"] = 20  # period for checking playlist length
    conf["EVENT_TICK_SECS"] = 30  # period for scheduling events
    conf["DEBUG"] = False
    conf["LOG_CONFIG"] = False
    conf["TMPDIR"] = os.getenv("TMPDIR", "/tmp/")
    conf["FILE_PATH_SUGGESTION"] = ()  # tuple of paths
    # UI_CALENDAR_FREQUENCY_THRESHOLD"] has been removed
    # use UI_CALENDAR_OCCURRENCIES_THRESHOLD instead
    conf["UI_CALENDAR_FREQUENCY_THRESHOLD"] = 4 * 60 * 60  # 4 hours
    conf["UI_CALENDAR_OCCURRENCIES_THRESHOLD"] = 40
    conf["UI_CALENDAR_DATE_FMT"] = "medium"
    conf["EVENT_FILTERS"] = []
    conf["HTTP_PORT"] = 5000
    conf["HOME_URL"] = "/db/calendar"
    conf.update(from_envvars(prefix=prefix))
    return conf


def from_envvars(prefix=None, envvars=None, as_json=True):
    """Load environment variables in a dictionary

    Values are parsed as JSON. If parsing fails with a ValueError,
    values are instead used as verbatim strings.

    :param prefix: If ``None`` is passed as envvars, all variables from
                   ``environ`` starting with this prefix are imported. The
                   prefix is stripped upon import.
    :param envvars: A dictionary of mappings of environment-variable-names
                    to Flask configuration names. If a list is passed
                    instead, names are mapped 1:1. If ``None``, see prefix
                    argument.
    :param as_json: If False, values will not be parsed as JSON first.
    """
    conf = {}
    if prefix is None and envvars is None:
        raise RuntimeError("Must either give prefix or envvars argument")

    # if it's a list, convert to dict
    if isinstance(envvars, list):
        envvars = {k: None for k in envvars}

    if not envvars:
        envvars = {
            k: k[len(prefix) :] for k in os.environ.keys() if k.startswith(prefix)
        }

    for env_name, name in envvars.items():
        if name is None:
            name = env_name

        if env_name not in os.environ:
            continue

        if as_json:
            try:
                conf[name] = json.loads(os.environ[env_name])
            except ValueError:
                conf[name] = os.environ[env_name]
        else:
            conf[name] = os.environ[env_name]

    return conf



"""
main module to read and get informations about alarms
"""
from __future__ import print_function
import sys
from datetime import datetime
import argparse
import json
from .entrypoints_utils import get_one_entrypoint
from logging import getLogger

log = getLogger("timegen")


def get_timegenerator(kind):
    """Messes with entrypoints to return an timegenerator function"""
    return get_one_entrypoint("larigira.timegenerators", kind)


def get_parser():
    parser = argparse.ArgumentParser(
        description='Generate "ring times" from a timespec'
    )
    parser.add_argument(
        "timespec",
        metavar="TIMESPEC",
        type=str,
        nargs=1,
        help="filename for timespec, formatted in json",
    )
    parser.add_argument(
        "--now",
        metavar="NOW",
        type=int,
        nargs=1,
        default=None,
        help='Set a different "time", in unix epoch',
    )
    parser.add_argument(
        "--howmany",
        metavar="N",
        type=int,
        nargs=1,
        default=[1],
        help='Set a different "time", in unix epoch',
    )
    return parser


def read_spec(fname):
    try:
        if fname == "-":
            return json.load(sys.stdin)
        with open(fname) as buf:
            return json.load(buf)
    except ValueError:
        sys.stderr.write("Error: invalid JSON\n")
        sys.exit(1)


def check_spec(spec):
    if "kind" not in spec:
        yield "Missing field 'kind'"


def timegenerate(spec, now=None, howmany=1):
    Alarm = get_timegenerator(spec["kind"])
    generator = Alarm(spec)
    if now is not None:
        if type(now) is not datetime:
            now = datetime.fromtimestamp(now)
    for _ in range(howmany):
        now = generator.next_ring(current_time=now)
        yield now


def main():
    """Main function for the "larigira-timegen" executable"""
    args = get_parser().parse_args()
    spec = read_spec(args.timespec[0])
    errors = tuple(check_spec(spec))
    if errors:
        log.error("Errors in timespec")
        for err in errors:
            sys.stderr.write("Error: {}\n".format(err))
        sys.exit(1)
    now = None if args.now is None else args.now.pop()
    howmany = None if args.howmany is None else args.howmany.pop()
    for time in timegenerate(spec, now=now, howmany=howmany):
        print(time)



from __future__ import print_function

import logging
from datetime import datetime, timedelta

import gevent
from gevent import monkey
from gevent.queue import Queue

from .audiogen import audiogenerate
from .db import EventModel
from .eventutils import ParentedLet, Timer
from .timegen import timegenerate

monkey.patch_all(subprocess=True)



logging.getLogger("mpd").setLevel(logging.WARNING)


class Monitor(ParentedLet):
    """
    Manages timegenerators and audiogenerators for DB events
    
    The mechanism is partially based on ticks, partially on scheduled actions.
    Ticks are emitted periodically; at every tick, :func:`on_tick
    <larigira.event.Monitor.on_tick>` checks if any event is "near enough". If
    an event is near enough, it is ":func:`scheduled
    <larigira.event.Monitor.schedule>`": a greenlet is run which will wait for
    the right time, then generate the audio, then submit to Controller.

    The tick mechanism allows for events to be changed on disk: if everything
    was scheduled immediately, no further changes would be possible.
    The scheduling mechanism allows for more precision, catching exactly the
    right time. Being accurate only with ticks would have required very
    frequent ticks, which is cpu-intensive.
    """

    def __init__(self, parent_queue, conf):
        ParentedLet.__init__(self, parent_queue)
        self.log = logging.getLogger(self.__class__.__name__)
        self.running = {}
        self.conf = conf
        self.q = Queue()
        self.model = EventModel(self.conf["DB_URI"])
        self.ticker = Timer(int(self.conf["EVENT_TICK_SECS"]) * 1000, self.q)

    def _alarm_missing_time(self, timespec):
        now = datetime.now() + timedelta(seconds=self.conf["CACHING_TIME"])
        try:
            when = next(timegenerate(timespec, now=now))
        except:
            logging.exception(
                "Could not generate " "an alarm from timespec %s", timespec
            )
        if when is None:
            # expired
            return None
        delta = (when - now).total_seconds()
        assert delta > 0
        return delta

    def on_tick(self):
        """
        this is called every EVENT_TICK_SECS.
        Checks every event in the DB (which might be slightly CPU-intensive, so
        it is advisable to run it in its own greenlet); if the event is "near
        enough", schedule it; if it is too far, or already expired, ignore it.
        """
        self.model.reload()
        for alarm in self.model.get_all_alarms():
            actions = list(self.model.get_actions_by_alarm(alarm))
            if alarm.eid in self.running:
                continue
            delta = self._alarm_missing_time(alarm)
            # why this 2*EVENT_TICK_SECS? EVENT_TICK_SECS would be enough,
            # but it is "tricky"; any small delay would cause the event to be
            # missed
            if delta is None:
                # this is way too much logging! we need more levels!
                # self.log.debug(
                #    "Skipping event %s: will never ring", alarm.get("nick", alarm.eid)
                # )
                pass
            elif delta <= 2 * self.conf["EVENT_TICK_SECS"]:
                self.log.debug(
                    "Scheduling event %s (%ds) => %s",
                    alarm.get("nick", alarm.eid),
                    delta,
                    [a.get("nick", a.eid) for a in actions],
                )
                self.schedule(alarm, actions, delta)
            else:
                self.log.debug(
                    "Skipping event %s too far (%ds)",
                    alarm.get("nick", alarm.eid),
                    delta,
                )

    def schedule(self, timespec, audiospecs, delta=None):
        """
        prepare an event to be run at a specified time with the specified
        actions; the DB won't be read anymore after this call.

        This means that this call should not be done too early, or any update
        to the DB will be ignored.
        """
        if delta is None:
            delta = self._alarm_missing_time(timespec)

        audiogen = gevent.spawn_later(delta, self.process_action, timespec, audiospecs)
        audiogen.parent_greenlet = self
        audiogen.doc = 'Will wait {} seconds, then generate audio "{}"'.format(
            delta, ",".join(aspec.get("nick", "") for aspec in audiospecs)
        )
        self.running[timespec.eid] = {
            "greenlet": audiogen,
            "running_time": datetime.now() + timedelta(seconds=delta),
            "timespec": timespec,
            "audiospecs": audiospecs,
        }

    def process_action(self, timespec, audiospecs):
        """Generate audio and submit it to Controller"""
        if timespec.eid in self.running:
            del self.running[timespec.eid]
        else:
            self.log.warning(
                "Timespec %s completed but not in running "
                "registry; this is most likely a bug",
                timespec.get("nick", timespec.eid),
            )
        uris = []
        for audiospec in audiospecs:
            try:
                uris.extend(audiogenerate(audiospec))
            except Exception as exc:
                self.log.error(
                    "audiogenerate for <%s> failed; reason: %s",
                    str(audiospec),
                    str(exc),
                )
        self.send_to_parent(
            "uris_enqueue",
            dict(
                uris=uris,
                timespec=timespec,
                audiospecs=audiospecs,
                aids=[a.eid for a in audiospecs],
            ),
        )

    def _run(self):
        self.ticker.start()
        gevent.spawn(self.on_tick)
        while True:
            value = self.q.get()
            kind = value["kind"]
            if kind in ("forcetick", "timer"):
                gevent.spawn(self.on_tick)
            else:
                self.log.warning("Unknown message: %s", str(value))



from flask_wtf import Form
from wtforms import StringField, validators, SubmitField

from larigira.formutils import AutocompleteStringField


class StaticAudioForm(Form):
    nick = StringField(
        "Audio nick",
        validators=[validators.required()],
        description="A simple name to recognize this audio",
    )
    path = AutocompleteStringField(
        "dl-suggested-files",
        "Path",
        validators=[validators.required()],
        description="Full path to audio file",
    )
    submit = SubmitField("Submit")

    def populate_from_audiospec(self, audiospec):
        if "nick" in audiospec:
            self.nick.data = audiospec["nick"]
        if "paths" in audiospec:
            self.path.data = audiospec["paths"][0]


def staticaudio_receive(form):
    return {"kind": "static", "nick": form.nick.data, "paths": [form.path.data]}



from flask_wtf import Form
from wtforms import StringField, validators, SubmitField


class AudioForm(Form):
    nick = StringField(
        "Audio nick",
        validators=[validators.required()],
        description="A simple name to recognize this audio",
    )
    urls = StringField(
        "URLs",
        validators=[validators.required()],
        description="URL of the file to download",
    )
    submit = SubmitField("Submit")

    def populate_from_audiospec(self, audiospec):
        if "nick" in audiospec:
            self.nick.data = audiospec["nick"]
        if "urls" in audiospec:
            self.urls.data = ";".join(audiospec["urls"])


def audio_receive(form):
    return {"kind": "http", "nick": form.nick.data, "urls": form.urls.data.split(";")}



"""
This module is for the main application logic
"""
from __future__ import print_function

import logging
import logging.config
import os
import signal
import subprocess
import sys
import tempfile
from time import sleep

import gevent
from gevent import monkey
from gevent.pywsgi import WSGIServer

from .config import get_conf
from .mpc import Controller, get_mpd_client
from .rpc import create_app

monkey.patch_all(subprocess=True)


def on_main_crash(*args, **kwargs):
    print('A crash occurred in "main" greenlet. Aborting...')
    sys.exit(1)




class Larigira(object):
    def __init__(self):

        self.log = logging.getLogger("larigira")
        self.conf = get_conf()
        self.controller = Controller(self.conf)
        self.controller.link_exception(on_main_crash)
        self.http_server = WSGIServer(
            ("", int(self.conf["HTTP_PORT"])), create_app(self.controller.q, self)
        )

    def start(self):
        self.controller.start()
        self.http_server.start()


def sd_notify(ready=False, status=None):
    args = ["systemd-notify"]
    if ready:
        args.append("--ready")
    if status is not None:
        args.append("--status")
        args.append(status)
    try:
        subprocess.check_call(args)
    except:
        pass


def main():
    tempfile.tempdir = os.environ["TMPDIR"] = os.path.join(
        os.getenv("TMPDIR", "/tmp"), "larigira.%d" % os.getuid()
    )
    if not os.path.isdir(os.environ["TMPDIR"]):
        os.makedirs(os.environ["TMPDIR"])

    if get_conf()["LOG_CONFIG"]:
        logging.config.fileConfig(
            get_conf()["LOG_CONFIG"], disable_existing_loggers=True
        )
    else:
        log_format = "%(asctime)s|%(levelname)s[%(name)s:%(lineno)d] %(message)s"
        logging.basicConfig(
            level=logging.DEBUG if get_conf()["DEBUG"] else logging.INFO,
            format=log_format,
            datefmt="%H:%M:%S",
        )
    logging.addLevelName(9, "DEBUGV")

    def debugv(self, message, *args, **kws):
        if self.isEnabledFor(9):
            self._log(9, message, args, **kws)

    logging.Logger.debugv = debugv

    if get_conf()["MPD_WAIT_START"]:

        while True:
            try:
                get_mpd_client(get_conf())
            except Exception as exc:
                print("exc", exc, file=sys.stderr)
                logging.debug(
                    "Could not connect to MPD at (%s,%s), waiting",
                    get_conf()["MPD_HOST"],
                    get_conf()["MPD_PORT"],
                )
                sd_notify(status="Waiting MPD connection")
                sleep(int(get_conf()["MPD_WAIT_START_RETRYSECS"]))
            else:
                logging.info("MPD ready!")
                sd_notify(ready=True, status="Ready")
                break

    larigira = Larigira()
    larigira.start()

    def sig(*args):
        print("invoked sig", args)
        larigira.controller.q.put(dict(kind="signal", args=args))

    for signum in (signal.SIGHUP, signal.SIGALRM):
        gevent.signal(signum, sig, signum)
    gevent.wait()


if __name__ == "__main__":
    main()



from logging import getLogger

log = getLogger("timeform")
from .entrypoints_utils import get_one_entrypoint


def get_timeform(kind):
    """Messes with entrypoints to return a TimeForm"""
    for group in ("larigira.timeform_create", "larigira.timeform_receive"):
        yield get_one_entrypoint(group, kind)


def get_audioform(kind):
    """Messes with entrypoints to return a AudioForm"""
    for group in ("larigira.audioform_create", "larigira.audioform_receive"):
        yield get_one_entrypoint(group, kind)



import logging
from datetime import datetime, timedelta

from croniter import croniter

from .timegen_every import Alarm

log = logging.getLogger("time-cron")


class CronAlarm(Alarm):

    description = "Frequency specified by cron-like format. nerds preferred"

    def __init__(self, obj):
        super().__init__()

        self.cron_format = obj["cron_format"]
        if "exclude" in obj:
            if type(obj["exclude"]) is str:
                self.exclude = [
                    line.strip() for line in obj["exclude"].split("\n") if line.strip()
                ]
            else:
                self.exclude = [excl for excl in obj["exclude"] if excl.strip()]
        else:
            self.exclude = []
        if not croniter.is_valid(self.cron_format):
            raise ValueError("Invalid cron_format: `%s`" % self.cron_format)
        for exclude in self.exclude:
            if not croniter.is_valid(exclude):
                raise ValueError("Invalid exclude: `%s`" % exclude)

    def is_excluded(self, dt):
        base = dt - timedelta(seconds=1)
        for exclude in self.exclude:
            nt = croniter(exclude, base).get_next(datetime)
            if nt == dt:
                return True
        return False

    def next_ring(self, current_time=None):
        if current_time is None:
            current_time = datetime.now()

        # cron granularity is to the minute
        # thus, doing 2000 attemps guarantees at least 32hours.
        # if your event is no more frequent than 10minutes, this is 13days
        for _ in range(2000):
            nt = croniter(self.cron_format, current_time).get_next(datetime)
            if not self.is_excluded(nt):
                return nt
            current_time = nt
        return None

    def has_ring(self, current_time=None):
        # cron specification has no possibility of being over
        return self.next_ring(current_time) is not None



"""
This module contains a flask blueprint for db administration stuff

Templates are self-contained in this directory
"""
from __future__ import print_function
import os
from datetime import datetime, timedelta, time
from collections import defaultdict
import mimetypes

from flask import (
    current_app,
    Blueprint,
    Response,
    render_template,
    jsonify,
    abort,
    request,
    redirect,
    url_for,
    flash,
)

from larigira.entrypoints_utils import get_avail_entrypoints
from larigira.audiogen import get_audiogenerator
from larigira.timegen_every import FrequencyAlarm
from larigira.timegen import get_timegenerator, timegenerate
from larigira import forms
from larigira.config import get_conf
from .suggestions import get_suggestions

db = Blueprint(
    "db",
    __name__,
    url_prefix=get_conf()["ROUTE_PREFIX"] + "/db",
    template_folder="templates",
)


def request_wants_json():
    best = request.accept_mimetypes.best_match(["application/json", "text/html"])
    return (
        best == "application/json"
        and request.accept_mimetypes[best] > request.accept_mimetypes["text/html"]
    )


def get_model():
    return current_app.larigira.controller.monitor.model


@db.route("/")
def home():
    return render_template("dbadmin_base.html")


@db.route("/list")
def events_list():
    model = current_app.larigira.controller.monitor.model
    alarms = tuple(model.get_all_alarms())
    events = [(alarm, model.get_actions_by_alarm(alarm)) for alarm in alarms]
    return render_template("list.html", events=events)


@db.route("/calendar")
def events_calendar():
    model = current_app.larigira.controller.monitor.model
    today = datetime.now().date()
    max_days = 30
    max_occurrences =  get_conf()["UI_CALENDAR_OCCURRENCIES_THRESHOLD"]
    # {date: {datetime: [(alarm1,actions1), (alarm2,actions2)]}}
    days = defaultdict(lambda: defaultdict(list))
    show_all = (request.args.get('all', '0') == '1')
    for alarm in model.get_all_alarms():
        actions = tuple(model.get_actions_by_alarm(alarm))
        if not actions:
            continue
        t = datetime.fromtimestamp(int(today.strftime("%s")))
        occurrences = [ t for t in timegenerate(alarm, now=t, howmany=max_occurrences +
                                        1) if t is not None and t <= datetime.combine(
                today + timedelta(days=max_days), time()
            )]
        if not occurrences:
            continue
        if not show_all and len(occurrences) > max_occurrences:
            continue
        for t in occurrences:
            days[t.date()][t].append((alarm, actions))

    # { weeknum: [day1, day2, day3] }
    weeks = defaultdict(list)
    for d in sorted(days.keys()):
        weeks[d.isocalendar()[:2]].append(d)

    return render_template("calendar.html", days=days, weeks=weeks,
                           show_all=show_all)


@db.route("/add/time")
def addtime():
    kinds = get_avail_entrypoints("larigira.timeform_create")

    def gen_info(gen):
        return dict(description=getattr(gen, "description", ""))

    info = {kind: gen_info(get_timegenerator(kind)) for kind in kinds}
    return render_template("add_time.html", kinds=kinds, info=info)


@db.route("/edit/time/<int:alarmid>", methods=["GET", "POST"])
def edit_time(alarmid):
    model = get_model()
    timespec = model.get_alarm_by_id(alarmid)
    kind = timespec["kind"]
    Form, receiver = tuple(forms.get_timeform(kind))
    form = Form()
    if request.method == "GET":
        form.populate_from_timespec(timespec)
    if request.method == "POST" and form.validate():
        data = receiver(form)
        model.update_alarm(alarmid, data)
        model.reload()
        return redirect(url_for("db.events_calendar", highlight="%d" % alarmid))
    return render_template(
        "add_time_kind.html", form=form, kind=kind, mode="edit", alarmid=alarmid
    )


def is_xhr():
    return request.headers.get("x-requested-with") is not None


@db.route("/add/time/<kind>", methods=["GET", "POST"])
def addtime_kind(kind):
    Form, receiver = tuple(forms.get_timeform(kind))
    form = Form(csrf_enabled=(not is_xhr()))
    if request.method == "POST":
        if form.validate():
            data = receiver(form)
            eid = get_model().add_alarm(data)
            if not is_xhr():
                return redirect(url_for("db.edit_event", alarmid=eid))
            else:
                resp = jsonify(alarmid=eid)
                resp.status_code = 201
                return resp
        elif is_xhr():
            resp = jsonify(errors=form.errors)
            resp.status_code = 400
            return resp

    return render_template("add_time_kind.html", form=form, kind=kind, mode="add")


@db.route("/add/audio")
def addaudio():
    kinds = get_avail_entrypoints("larigira.audioform_create")

    def gen_info(gen):
        return dict(description=getattr(gen, "description", ""))

    info = {kind: gen_info(get_audiogenerator(kind)) for kind in kinds}
    return render_template("add_audio.html", kinds=kinds, info=info)


@db.route("/add/audio/<kind>", methods=["GET", "POST"])
def addaudio_kind(kind):
    Form, receiver = tuple(forms.get_audioform(kind))
    form = Form(csrf_enabled=(not is_xhr()))
    if request.method == "POST":
        if form.validate():
            data = receiver(form)
            model = current_app.larigira.controller.monitor.model
            eid = model.add_action(data)
            return jsonify(dict(inserted=eid, data=data))
        else:
            resp = jsonify(errors=form.errors)
            resp.status_code = 400
            return resp

    return render_template(
        "add_audio_kind.html", form=form, kind=kind, suggestions=get_suggestions()
    )


@db.route("/edit/audio/<int:actionid>", methods=["GET", "POST"])
def edit_audio(actionid):
    model = get_model()
    audiospec = model.get_action_by_id(actionid)
    kind = audiospec["kind"]
    Form, receiver = tuple(forms.get_audioform(kind))
    form = Form()
    if request.method == "GET":
        form.populate_from_audiospec(audiospec)
    if request.method == "POST" and form.validate():
        data = receiver(form)
        model.update_action(actionid, data)
        model.reload()
        return redirect(url_for("db.events_list"))
    return render_template(
        "add_audio_kind.html",
        form=form,
        kind=kind,
        mode="edit",
        suggestions=get_suggestions(),
    )


@db.route("/edit/event/<alarmid>")
def edit_event(alarmid):
    model = current_app.larigira.controller.monitor.model
    alarm = model.get_alarm_by_id(int(alarmid))
    if alarm is None:
        abort(404)
    allactions = model.get_all_actions()
    actions = tuple(model.get_actions_by_alarm(alarm))
    return render_template(
        "edit_event.html",
        alarm=alarm,
        all_actions=allactions,
        actions=actions,
        routeprefix=get_conf()["ROUTE_PREFIX"],
    )


@db.route("/api/alarm/<alarmid>/actions", methods=["POST"])
def change_actions(alarmid):
    new_actions = request.form.getlist("actions[]")
    if new_actions is None:
        new_actions = []
    model = current_app.larigira.controller.monitor.model
    ret = model.update_alarm(
        int(alarmid), new_fields={"actions": [int(a) for a in new_actions]}
    )
    return jsonify(dict(updated=alarmid, ret=ret))


@db.route("/api/alarm/<int:alarmid>/delete", methods=["POST"])
def delete_alarm(alarmid):
    model = current_app.larigira.controller.monitor.model
    try:
        alarm = model.get_alarm_by_id(int(alarmid))
        print(alarm["nick"])
        model.delete_alarm(alarmid)
    except KeyError:
        abort(404)
    if request_wants_json():
        return jsonify(dict(deleted=alarmid))
    flash("Evento %d `%s` cancellato" % (alarmid, alarm["nick"]))
    return redirect(url_for("db.events_list"))


@db.route("/extra/<path:relpath>")
def static_custom(relpath):
    basepath = get_conf()["EXTRA_STATIC_PATH"]
    if not basepath:
        abort(405)
    fpath = os.path.join(basepath, relpath)
    print(basepath, fpath)
    if not os.path.isfile(fpath):
        abort(404, "File non trovato")
    mime, _encoding = mimetypes.guess_type(fpath)
    return Response(open(fpath, "rb").read(), mimetype=mime)



import logging
import random

from mpd import MPDClient

from .config import get_conf

log = logging.getLogger(__name__)


def generate_by_artist(spec):
    """choose HOWMANY random artists, and for each one choose a random song"""

    # TODO: support prefix
    spec.setdefault("howmany", 1)
    prefix = spec.get("prefix", "").rstrip("/")
    log.info("generating")
    conf = get_conf()
    c = MPDClient(use_unicode=True)
    c.connect(conf["MPD_HOST"], conf["MPD_PORT"])

    if prefix:
        artists = list({r["artist"] for r in c.listallinfo(prefix) if "artist" in r})
    else:
        artists = c.list("artist")
    if not artists:
        raise ValueError("no artists in your mpd database")
    for _ in range(spec["howmany"]):
        artist = random.choice(artists)  # pick one artist
        # pick one song from that artist
        artist_songs = (res["file"] for res in c.find("artist", artist))
        if prefix:
            artist_songs = [
                fname for fname in artist_songs if fname.startswith(prefix + "/")
            ]
        yield random.choice(list(artist_songs))



import os
import logging
import posixpath
from tempfile import mkstemp
import urllib.request
from urllib.parse import urlparse

log = logging.getLogger(__name__)


def put(url, destdir=None, copy=False):
    if url.split(":")[0] not in ("http", "https"):
        log.warning("Not a valid URL: %s", url)
        return None
    ext = url.split(".")[-1]
    if ext.lower() not in ("mp3", "ogg", "oga", "wma", "m4a"):
        log.warning('Invalid format (%s) for "%s"', ext, url)
        return None
    if not copy:
        return url
    fname = posixpath.basename(urlparse(url).path)
    # sanitize
    fname = "".join(c for c in fname if c.isalnum() or c in list("._-")).rstrip()
    tmp = mkstemp(suffix="." + ext, prefix="http-%s-" % fname, dir=destdir)
    os.close(tmp[0])
    log.info("downloading %s -> %s", url, tmp[1])
    fname, headers = urllib.request.urlretrieve(url, tmp[1])
    return "file://%s" % os.path.realpath(tmp[1])


def generate(spec):
    """
    resolves audiospec-static

    Recognized argument is  "paths" (list of static paths)
    """
    if "urls" not in spec:
        raise ValueError("Malformed audiospec: missing 'paths'")

    for url in spec["urls"]:
        ret = put(url, copy=True)
        if ret is None:
            continue
        yield ret


generate.description = "Fetch audio from an URL"



from tinydb import TinyDB


class EventModel(object):
    def __init__(self, uri):
        self.uri = uri
        self.db = None
        self.reload()

    def reload(self):
        if self.db is not None:
            self.db.close()
        self.db = TinyDB(self.uri, indent=2)
        self._actions = self.db.table("actions")
        self._alarms = self.db.table("alarms")

    def get_action_by_id(self, action_id):
        return self._actions.get(eid=action_id)

    def get_alarm_by_id(self, alarm_id):
        return self._alarms.get(eid=alarm_id)

    def get_actions_by_alarm(self, alarm):
        for action_id in alarm.get("actions", []):
            action = self.get_action_by_id(action_id)
            if action is None:
                continue
            yield action

    def get_all_alarms(self):
        return self._alarms.all()

    def get_all_actions(self):
        return self._actions.all()

    def get_all_alarms_expanded(self):
        for alarm in self.get_all_alarms():
            for action in self.get_actions_by_alarm(alarm):
                yield alarm, action

    def add_event(self, alarm, actions):
        action_ids = [self.add_action(a) for a in actions]
        alarm["actions"] = action_ids
        return self._alarms.insert(alarm)

    def add_action(self, action):
        return self._actions.insert(action)

    def add_alarm(self, alarm):
        return self.add_event(alarm, [])

    def update_alarm(self, alarmid, new_fields={}):
        return self._alarms.update(new_fields, eids=[alarmid])

    def update_action(self, actionid, new_fields={}):
        return self._actions.update(new_fields, eids=[actionid])

    def delete_alarm(self, alarmid):
        return self._alarms.remove(eids=[alarmid])

    def delete_action(self, actionid):
        return self._actions.remove(eids=[actionid])



import pytest

from .unused import UnusedCleaner
from .config import get_conf


@pytest.fixture
def unusedcleaner():
    return UnusedCleaner(get_conf(prefix="LARIGIRATEST_"))


# this test suite heavily assumes that TMPDIR == /tmp/, which is the default
# indeed.  However, the code does not rely on this assumption.


def test_watch_file(unusedcleaner):
    # despite not existing, the file is added
    unusedcleaner.watch("file:///tmp/gnam")
    assert len(unusedcleaner.waiting_removal_files) == 1
    assert list(unusedcleaner.waiting_removal_files)[0] == "/tmp/gnam"


def test_watch_path_error(unusedcleaner):
    """paths are not valid thing to watch. URIs only, thanks"""
    unusedcleaner.watch("/tmp/foo")
    assert len(unusedcleaner.waiting_removal_files) == 0


def test_watch_notmp_error(unusedcleaner):
    """Files not in TMPDIR are not added"""
    unusedcleaner.watch("file:///not/in/tmp")
    assert len(unusedcleaner.waiting_removal_files) == 0



import logging

from datetime import datetime
from pytimeparse.timeparse import timeparse

from flask_wtf import Form
from wtforms import (
    StringField,
    validators,
    SubmitField,
    SelectMultipleField,
    ValidationError,
)

from larigira.formutils import EasyDateTimeField

log = logging.getLogger(__name__)


class SingleAlarmForm(Form):
    nick = StringField(
        "Alarm nick",
        validators=[validators.required()],
        description="A simple name to recognize this alarm",
    )
    dt = EasyDateTimeField(
        "Date and time",
        validators=[validators.required()],
        description="Date to ring on, expressed as " "2000-12-31T13:42:00",
    )
    submit = SubmitField("Submit")

    def populate_from_timespec(self, timespec):
        if "nick" in timespec:
            self.nick.data = timespec["nick"]
        if "timestamp" in timespec:
            self.dt.data = datetime.fromtimestamp(timespec["timestamp"])


def singlealarm_receive(form):
    return {
        "kind": "single",
        "nick": form.nick.data,
        "timestamp": int(form.dt.data.strftime("%s")),
    }


class FrequencyAlarmForm(Form):
    nick = StringField(
        "Alarm nick",
        validators=[validators.required()],
        description="A simple name to recognize this alarm",
    )
    interval = StringField(
        "Frequency",
        validators=[validators.required()],
        description="in seconds, or human-readable " "(like 9w3d12h)",
    )
    start = EasyDateTimeField(
        "Start date and time",
        validators=[validators.optional()],
        description="Before this, no alarm will ring. "
        "Expressed as YYYY-MM-DDTHH:MM:SS. If omitted, "
        "the alarm will always ring",
    )
    end = EasyDateTimeField(
        "End date and time",
        validators=[validators.optional()],
        description="After this, no alarm will ring. "
        "Expressed as YYYY-MM-DDTHH:MM:SS. If omitted, "
        "the alarm will always ring",
    )
    weekdays = SelectMultipleField(
        "Days on which the alarm should be played",
        choices=[
            ("1", "Monday"),
            ("2", "Tuesday"),
            ("3", "Wednesday"),
            ("4", "Thursday"),
            ("5", "Friday"),
            ("6", "Saturday"),
            ("7", "Sunday"),
        ],
        default=list("1234567"),
        validators=[validators.required()],
        description="The alarm will ring only on " "selected weekdays",
    )
    submit = SubmitField("Submit")

    def populate_from_timespec(self, timespec):
        if "nick" in timespec:
            self.nick.data = timespec["nick"]
        if "start" in timespec:
            self.start.data = datetime.fromtimestamp(timespec["start"])
        if "end" in timespec:
            self.end.data = datetime.fromtimestamp(timespec["end"])
        if "weekdays" in timespec:
            self.weekdays.data = timespec["weekdays"]
        else:
            self.weekdays.data = list("1234567")
        self.interval.data = timespec["interval"]

    def validate_interval(self, field):
        try:
            int(field.data)
        except ValueError:
            if timeparse(field.data) is None:
                raise ValidationError(
                    "interval must either be a number "
                    "(in seconds) or a human-readable "
                    "string like '1h2m'  or '1d12h'"
                )


def frequencyalarm_receive(form):
    obj = {
        "kind": "frequency",
        "nick": form.nick.data,
        "interval": form.interval.data,
        "weekdays": form.weekdays.data,
    }
    if form.start.data:
        obj["start"] = int(form.start.data.strftime("%s"))
    else:
        obj["start"] = 0
    if form.end.data:
        obj["end"] = int(form.end.data.strftime("%s"))
    return obj



'''
This module provides a simple to use way for method (de)serialization focused on
keeping state across restarts. Its simple API will allow you to have internal state kept on filesystem and
recovered.
'''

# TODO: make it a singleton
class StateKeeper:
    @classmethod
    def register(cls, instance_name, obj):
        # TODO: deserialize (if you can)
        # TODO: run sth that will dump it every X
        # TODO: attach a handler to dump upon exit
        pass




import os

from flask import current_app

from larigira.config import get_conf
from larigira.fsutils import scan_dir_audio


def get_suggested_files():
    if not get_conf()["FILE_PATH_SUGGESTION"]:
        return []
    if current_app.cache.has("dbadmin.get_suggested_files"):
        return current_app.cache.get("dbadmin.get_suggested_files")
    current_app.logger.debug("get_suggested_files MISS in cache")
    files = []
    for path in get_conf()["FILE_PATH_SUGGESTION"]:
        if not os.path.isdir(path):
            current_app.logger.warning("Invalid suggestion path: %s" % path)
            continue
        pathfiles = scan_dir_audio(path)
        files.extend(pathfiles)
    current_app.logger.debug("Suggested files: %s" % ", ".join(files))

    current_app.cache.set(
        "dbadmin.get_suggested_files", files, timeout=600
    )  # ten minutes
    return files


def get_suggested_dirs():
    dirset = set()
    for f in get_suggested_files():
        dirpath = os.path.dirname(f)
        while dirpath:
            if dirpath in dirset:
                break
            dirset.add(dirpath)
            dirpath = os.path.dirname(dirpath)

    return list(dirset)


def get_suggested_scripts():
    base = get_conf()["SCRIPTS_PATH"]
    if not base or not os.path.isdir(base):
        return []
    fnames = [
        f
        for f in os.listdir(base)
        if os.access(os.path.join(base, f), os.R_OK | os.X_OK)
    ]
    return fnames


def get_suggestions():
    files = get_suggested_files()
    if len(files) > 200:
        current_app.logger.warning("Too many suggested files, cropping")
        files = files[:200]
    return dict(files=files, dirs=get_suggested_dirs(), scripts=get_suggested_scripts())



import os
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as buf:
        return buf.read()


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name="larigira",
    version="1.3.2",
    description="A radio automation based on MPD",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    author="boyska",
    author_email="piuttosto@logorroici.org",
    license="AGPL",
    packages=["larigira", "larigira.dbadmin", "larigira.filters"],
    install_requires=[
        "Babel==2.6.0",
        "Flask-Babel==1.0.0",
        "pyxdg",
        "gevent",
        "flask-bootstrap",
        "python-mpd2",
        "wtforms",
        "Flask-WTF",
        "flask==0.11",
        "pytimeparse",
        "croniter==0.3.29",
        "werkzeug==0.14.1",
        "cachelib==0.1",
        "tinydb",
    ],
    tests_require=["pytest-timeout==1.0", "py>=1.4.29", "pytest==3.0"],
    python_requires=">=3.5",
    extras_require={"percentwait": ["mutagen"]},
    cmdclass={"test": PyTest},
    zip_safe=False,
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "larigira=larigira.larigira:main",
            "larigira-timegen=larigira.timegen:main",
            "larigira-audiogen=larigira.audiogen:main",
            "larigira-dbmanage=larigira.event_manage:main",
        ],
        "larigira.audiogenerators": [
            "mpd = larigira.audiogen_mpdrandom:generate_by_artist",
            "static = larigira.audiogen_static:generate",
            "http = larigira.audiogen_http:generate",
            "randomdir = larigira.audiogen_randomdir:generate",
            "mostrecent = larigira.audiogen_mostrecent:generate",
            "script = larigira.audiogen_script:generate",
        ],
        "larigira.timegenerators": [
            "frequency = larigira.timegen_every:FrequencyAlarm",
            "single = larigira.timegen_every:SingleAlarm",
            "cron = larigira.timegen_cron:CronAlarm",
        ],
        "larigira.timeform_create": [
            "single = larigira.timeform_base:SingleAlarmForm",
            "frequency = larigira.timeform_base:FrequencyAlarmForm",
            "cron = larigira.timeform_cron:CronAlarmForm",
        ],
        "larigira.timeform_receive": [
            "single = larigira.timeform_base:singlealarm_receive",
            "frequency = larigira.timeform_base:frequencyalarm_receive",
            "cron = larigira.timeform_cron:cronalarm_receive",
        ],
        "larigira.audioform_create": [
            "static = larigira.audioform_static:StaticAudioForm",
            "http = larigira.audioform_http:AudioForm",
            "script = larigira.audioform_script:ScriptAudioForm",
            "randomdir = larigira.audioform_randomdir:Form",
            "mostrecent = larigira.audioform_mostrecent:AudioForm",
        ],
        "larigira.audioform_receive": [
            "static = larigira.audioform_static:staticaudio_receive",
            "http = larigira.audioform_http:audio_receive",
            "script = larigira.audioform_script:scriptaudio_receive",
            "randomdir = larigira.audioform_randomdir:receive",
            "mostrecent = larigira.audioform_mostrecent:audio_receive",
        ],
        "larigira.eventfilter": [
            "maxwait = larigira.filters:maxwait",
            "percentwait = larigira.filters:percentwait",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3",
    ],
)



import gc
import logging
from copy import deepcopy

from flask import (Blueprint, Flask, abort, current_app, jsonify, redirect,
                   render_template, request)
from flask.ext.babel import Babel
from flask_bootstrap import Bootstrap

from cachelib import SimpleCache
from greenlet import greenlet

from .config import get_conf
from .dbadmin import db

rpc = Blueprint("rpc", __name__, url_prefix=get_conf()["ROUTE_PREFIX"] + "/api")
viewui = Blueprint(
    "view",
    __name__,
    url_prefix=get_conf()["ROUTE_PREFIX"] + "/view",
    template_folder="templates",
)


def send_to_parent(kind, *args):
    """similar to the behaviour of a ParentedLet"""
    if not hasattr(current_app, "queue"):
        logging.debug("no parent queue; aborting send")
        return
    msg = {
        "emitter": current_app._get_current_object(),
        "class": current_app._get_current_object().__class__.__name__,
        "kind": kind,
        "args": args,
    }
    current_app.queue.put(msg)


@rpc.route("/")
def rpc_index():
    rules = list(current_app.url_map.iter_rules())
    return jsonify({"rules": [r.rule for r in rules if r.rule.startswith("/api")]})


@rpc.route("/refresh")
def rpc_refresh():
    send_to_parent("refresh")
    return jsonify(dict(status="ok"))


@rpc.route("/audiospec", methods=["GET"])
def get_audiospec():
    return jsonify(current_app.larigira.controller.player.continous_audiospec)


@rpc.route("/audiospec", methods=["PUT"])
def change_audiospec():
    player = current_app.larigira.controller.player
    if request.json is None:
        abort(400, "Must send application/json data")
    if "spec" not in request.json or type(request.json["spec"]) is not dict:
        abort(400, "Object must have a key 'spec' whose value is an object")
    player.continous_audiospec = request.json["spec"]
    if "kind" not in request.json["spec"]:
        abort(400, "invalid audiospec")
    return jsonify(player.continous_audiospec)


@rpc.route("/audiospec", methods=["DELETE"])
def reset_audiospec():
    player = current_app.larigira.controller.player
    player.continous_audiospec = None
    return jsonify(player.continous_audiospec)


@rpc.route("/eventsenabled/toggle", methods=["POST"])
def toggle_events_enabled():
    status = current_app.larigira.controller.player.events_enabled
    current_app.larigira.controller.player.events_enabled = not status
    return jsonify(dict(events_enabled=not status))


@rpc.route("/eventsenabled", methods=["GET"])
def get_events_enabled():
    status = current_app.larigira.controller.player.events_enabled
    return jsonify(dict(events_enabled=status))


@rpc.route("/eventsenabled", methods=["PUT"])
def set_events_enabled():
    player = current_app.larigira.controller.player
    if request.json is None:
        abort(400, "Must send application/json data")
    if type(request.json) is not bool:
        abort(400, "Content must be a JSON boolean")
    player.events_enabled = request.json
    return jsonify(dict(events_enabled=request.json))


def get_scheduled_audiogen():
    larigira = current_app.larigira
    running = larigira.controller.monitor.running
    events = {t: {} for t in running.keys()}
    for timespec_eid in events:
        orig_info = running[timespec_eid]
        info = events[timespec_eid]
        info["running_time"] = orig_info["running_time"].isoformat()
        info["audiospecs"] = orig_info["audiospecs"]
        info["timespec"] = orig_info["timespec"]
        info["timespec"]["actions"] = {
            aid: spec
            for aid, spec in zip(info["timespec"]["actions"], info["audiospecs"])
        }
        info["greenlet"] = hex(id(orig_info["greenlet"]))
    return events


@viewui.route("/status/running")
def ui_wip():
    audiogens = get_scheduled_audiogen()
    return render_template(
        "running.html",
        audiogens=sorted(audiogens.items(), key=lambda x: x[1]["running_time"]),
    )


@rpc.route("/debug/running")
def rpc_wip():
    def treeify(flat):
        roots = [obid for obid in flat if flat[obid]["parent"] not in flat]
        tree = deepcopy(flat)
        for obid in tree:
            tree[obid]["children"] = {}

        to_remove = []
        for obid in tree:
            if obid in roots:
                continue
            if obid not in tree:
                current_app.logger.warning("How strange, {} not in tree".format(obid))
                continue
            tree[tree[obid]["parent"]]["children"][obid] = tree[obid]
            to_remove.append(obid)

        for obid in to_remove:
            del tree[obid]
        return tree

    greenlets = {}
    for ob in filter(lambda obj: isinstance(obj, greenlet), gc.get_objects()):
        objrepr = {"repr": repr(ob), "class": ob.__class__.__name__}
        if hasattr(ob, "parent_greenlet") and ob.parent_greenlet is not None:
            objrepr["parent"] = hex(id(ob.parent_greenlet))
        else:
            objrepr["parent"] = hex(id(ob.parent)) if ob.parent is not None else None
        if hasattr(ob, "doc"):
            objrepr["doc"] = ob.doc.split("\n")[0]
        elif ob.__doc__:
            objrepr["doc"] = ob.__doc__.split("\n")[0]

        greenlets[hex(id(ob))] = objrepr

    # TODO: make it a tree

    return jsonify(
        dict(
            greenlets=greenlets,
            greenlets_tree=treeify(greenlets),
            audiogens=get_scheduled_audiogen(),
        )
    )


def babel_get_locale():
    try:
        if request.accept_languages:
            return request.accept_languages[0][0]
    finally:
        return None


def create_app(queue, larigira):
    app = Flask("larigira", static_url_path=get_conf()["ROUTE_PREFIX"] + "/static")
    app.config.update(get_conf())
    Bootstrap(app)
    babel = Babel(app)
    babel.localeselector(babel_get_locale)
    app.register_blueprint(rpc)
    app.register_blueprint(viewui)
    app.register_blueprint(db)
    app.route(get_conf()["ROUTE_PREFIX"] + "/")(
        lambda: redirect(get_conf()["ROUTE_PREFIX"] + get_conf()["HOME_URL"])
    )
    app.queue = queue
    app.larigira = larigira
    app.cache = SimpleCache()
    return app



from __future__ import print_function
import logging
import signal
from pkg_resources import iter_entry_points

import gevent
from gevent.queue import Queue
import mpd

from .event import Monitor
from .eventutils import ParentedLet, Timer
from .audiogen import audiogenerate
from .unused import UnusedCleaner
from .entrypoints_utils import get_avail_entrypoints


def get_mpd_client(conf):
    client = mpd.MPDClient(use_unicode=True)
    client.connect(conf["MPD_HOST"], conf["MPD_PORT"])

    return client


class MPDWatcher(ParentedLet):
    """
    MPDWatcher notifies parent about any mpd event
    """

    def __init__(self, queue, conf, client=None):
        ParentedLet.__init__(self, queue)
        self.log = logging.getLogger(self.__class__.__name__)
        self.conf = conf
        self.client = client  # assume it is already connected, or None

    def refresh_client(self):
        self.client = get_mpd_client(self.conf)

    def do_business(self):
        first_after_connection = True
        while True:
            try:
                if self.client is None:
                    self.refresh_client()
                if first_after_connection:
                    yield ("mpc", "connect")

                status = self.client.idle()[0]
            except (
                mpd.ConnectionError,
                ConnectionRefusedError,
                FileNotFoundError,
            ) as exc:
                self.log.warning(
                    "Connection to MPD failed (%s: %s)", exc.__class__.__name__, exc
                )
                self.client = None
                first_after_connection = True
                gevent.sleep(5)
                continue
            else:
                first_after_connection = False
                yield ("mpc", status)


class Player:
    """
    The player contains different mpd-related methods

    check_playlist determines whether the playlist is long enough and run audiogenerator accordingly

    enqueue receive audios that have been generated by Monitor and (if filters allow it) enqueue it to MPD playlist
    """

    def __init__(self, conf):
        self.conf = conf
        self.log = logging.getLogger(self.__class__.__name__)
        self.min_playlist_length = 10
        self.tmpcleaner = UnusedCleaner(conf)
        self._continous_audiospec = self.conf["CONTINOUS_AUDIOSPEC"]
        self.events_enabled = True

    def _get_mpd(self):
        mpd_client = mpd.MPDClient(use_unicode=True)
        try:
            mpd_client.connect(self.conf["MPD_HOST"], self.conf["MPD_PORT"])
        except (mpd.ConnectionError, ConnectionRefusedError, FileNotFoundError) as exc:
            self.log.warning(
                "Connection to MPD failed (%s: %s)", exc.__class__.__name__, exc
            )
            raise gevent.GreenletExit()
        return mpd_client

    @property
    def continous_audiospec(self):
        return self._continous_audiospec

    @continous_audiospec.setter
    def continous_audiospec(self, spec):
        self._continous_audiospec = (
            self.conf["CONTINOUS_AUDIOSPEC"] if spec is None else spec
        )

        def clear_everything_but_current_song():
            mpdc = self._get_mpd()
            current = mpdc.currentsong()
            pos = int(current.get("pos", 0))
            for song in mpdc.playlistid():
                if int(song["pos"]) != pos:
                    mpdc.deleteid(song["id"])

        gevent.Greenlet.spawn(clear_everything_but_current_song)

    def check_playlist(self):
        mpd_client = self._get_mpd()
        songs = mpd_client.playlist()
        current = mpd_client.currentsong()
        pos = int(current.get("pos", 0)) + 1
        if (len(songs) - pos) >= self.min_playlist_length:
            return
        self.log.info("need to add new songs")
        picker = gevent.Greenlet(audiogenerate, self.continous_audiospec)

        def add(greenlet):
            uris = greenlet.value
            for uri in uris:
                assert type(uri) is str, type(uri)
                self.tmpcleaner.watch(uri.strip())
                mpd_client.add(uri.strip())

        picker.link_value(add)
        picker.start()

    def enqueue_filter(self, songs):
        eventfilters = self.conf["EVENT_FILTERS"]
        if not eventfilters:
            return True, ""
        availfilters = get_avail_entrypoints("larigira.eventfilter")
        if len([ef for ef in eventfilters if ef in availfilters]) == 0:
            return True, ""
        mpdc = self._get_mpd()
        status = mpdc.status()
        ctx = {"playlist": mpdc.playlist(), "status": status, "durations": []}
        for entrypoint in iter_entry_points("larigira.eventfilter"):
            if entrypoint.name in eventfilters:
                ef = entrypoint.load()
                try:
                    ret = ef(songs=songs, context=ctx, conf=self.conf)
                except ImportError as exc:
                    self.log.warn("Filter %s skipped: %s" % (entrypoint.name, exc))
                    continue
                if ret is None:  # bad behavior!
                    continue
                if type(ret) is bool:
                    reason = ""
                else:
                    ret, reason = ret
                reason = "Filtered by %s (%s)" % (entrypoint.name, reason)
                if ret is False:
                    return ret, reason
                else:
                    if reason:
                        self.log.debug('filter %s says ok: %s', entrypoint.name, reason)
        return True, "Passed through %s" % ",".join(availfilters)

    def enqueue(self, songs):
        assert type(songs) is dict
        assert "uris" in songs
        spec = [aspec.get("nick", aspec.eid) for aspec in songs["audiospecs"]]
        nicks = ",".join(
            (aspec.get("nick", aspec.eid) for aspec in songs["audiospecs"])
        )
        if not self.events_enabled:
            self.log.debug("Ignoring <%s> (events disabled)", nicks)
            return
        filterok, reason = self.enqueue_filter(songs)
        if not filterok:
            self.log.debug("Ignoring <%s>, filtered: %s", nicks, reason)
            # delete those files
            for uri in reversed(songs["uris"]):
                self.tmpcleaner.watch(uri.strip())
            return
        mpd_client = self._get_mpd()
        for uri in reversed(songs["uris"]):
            assert type(uri) is str
            self.log.info(
                "Adding %s to playlist (from <%s>:%s=%s)",
                uri,
                songs["timespec"].get("nick", ""),
                songs["aids"],
                spec,
            )
            insert_pos = (
                0
                if len(mpd_client.playlistid()) == 0
                else int(mpd_client.currentsong().get("pos", 0)) + 1
            )
            try:
                mpd_client.addid(uri, insert_pos)
            except mpd.CommandError:
                self.log.exception("Cannot insert song %s", uri)
            self.tmpcleaner.watch(uri.strip())


class Controller(gevent.Greenlet):
    def __init__(self, conf):
        gevent.Greenlet.__init__(self)
        self.log = logging.getLogger(self.__class__.__name__)
        self.conf = conf
        self.q = Queue()
        self.player = Player(self.conf)
        if "DB_URI" in self.conf:
            self.monitor = Monitor(self.q, self.conf)
            self.monitor.parent_greenlet = self
        else:
            self.monitor = None

    def _run(self):
        if self.monitor is not None:
            self.monitor.start()
        mw = MPDWatcher(self.q, self.conf, client=None)
        mw.parent_greenlet = self
        mw.start()
        t = Timer(int(self.conf["CHECK_SECS"]) * 1000, self.q)
        t.parent_greenlet = self
        t.start()
        # at the very start, run a check!
        gevent.Greenlet.spawn(self.player.check_playlist)
        while True:
            value = self.q.get()
            # emitter = value['emitter']
            kind = value["kind"]
            args = value["args"]
            if kind == "timer" or (
                kind == "mpc" and args[0] in ("player", "playlist", "connect")
            ):
                gevent.Greenlet.spawn(self.player.check_playlist)
                try:
                    self.player.tmpcleaner.check_playlist()
                except:
                    pass
            elif kind == "mpc":
                pass
            elif kind == "uris_enqueue":
                # TODO: uris_enqueue messages should be delivered directly to Player.enqueue
                #   probably we need a MPDEnqueuer that receives every uri we want to add
                try:
                    self.player.enqueue(args[0])
                except AssertionError:
                    raise
                except Exception:
                    self.log.exception(
                        "Error while adding to queue; " "bad audiogen output?"
                    )
            elif (kind == "signal" and args[0] == signal.SIGALRM) or kind == "refresh":
                # it's a tick!
                self.log.debug("Reload")
                self.monitor.q.put(dict(kind="forcetick"))
                gevent.Greenlet.spawn(self.player.check_playlist)
            else:
                self.log.warning("Unknown message: %s", str(value))



import os
import logging
import shutil
import random
from tempfile import mkstemp
from pathlib import Path

from larigira.fsutils import scan_dir_audio, shortname, is_audio

log = logging.getLogger(__name__)


def candidates(paths):
    c = set()
    for path in paths:
        if not path.exists():
            log.warning("Can't find requested path: %s", path)
            continue
        if path.is_file() and is_audio(str(path)):
            c.add(str(path))
        elif path.is_dir():
            c.update(scan_dir_audio(str(path)))
    return c


def generate(spec):
    """
    resolves audiospec-randomdir

    Recognized arguments:
        - paths [mandatory]    list of source paths
        - howmany [default=1]  number of audio files to pick
    """
    spec.setdefault("howmany", 1)
    for attr in ("paths",):
        if attr not in spec:
            raise ValueError("Malformed audiospec: missing '%s'" % attr)

    found_files = candidates([Path(p) for p in spec["paths"]])

    picked = random.sample(found_files, int(spec["howmany"]))

    nick = spec.get("nick", "")
    if not nick:
        if hasattr(spec, "eid"):
            nick = spec.eid
        else:
            nick = "NONICK"
    for path in picked:
        tmp = mkstemp(
            suffix=os.path.splitext(path)[-1],
            prefix="randomdir-%s-%s-" % (shortname(nick), shortname(path)),
        )
        os.close(tmp[0])
        shutil.copy(path, tmp[1])
        log.info("copying %s -> %s", path, os.path.basename(tmp[1]))
        yield "file://{}".format(tmp[1])


generate.description = "Picks random files from a specified directory"
